# keyword.py - $Keyword$ expansion for Mercurial
#
# Copyright 2007 Christian Ebert <blacktrash@gmx.net>
#
# This software may be used and distributed according to the terms
# of the GNU General Public License, incorporated herein by reference.
#
# $Id$
#
# Keyword expansion hack against the grain of a DSCM
#
# There are many good reasons why this is not needed in a distributed
# SCM, still it may be useful in very small projects based on single
# files (like LaTeX packages), that are mostly addressed to an audience
# not running a version control system.
#
# For in-depth discussion refer to
# <http://www.selenic.com/mercurial/wiki/index.cgi/KeywordPlan>.
#
# Keyword expansion is based on Mercurial's changeset template mappings.
#
# Binary files are not touched.
#
# Setup in hgrc:
#
#   [extensions]
#   # enable extension
#   keyword = /full/path/to/hgkw/keyword.py
#   # or, if script in canonical hgext folder:
#   # hgext.keyword =
#
# Files to act upon/ignore are specified in the [keyword] section.
# Customized keyword template mappings in the [keywordmaps] section.
#
# Run "hg help keyword" and "hg kwdemo" to get info on configuration.

'''keyword expansion in local repositories

This extension expands RCS/CVS-like or self-customized $Keywords$
in tracked text files selected by your configuration.

Keywords are only expanded in local repositories and not stored in
the change history. The mechanism can be regarded as a convenience
for the current user or for archive distribution.

Configuration is done in the [keyword] and [keywordmaps] sections
of hgrc files.

Example:

    [keyword]
    # expand keywords in every python file except those matching "x*"
    **.py =
    x*    = ignore

Note: the more specific you are in your filename patterns
      the less you lose speed in huge repos.

For [keywordmaps] template mapping and expansion demonstration and
control run "hg kwdemo".

An additional date template filter {date|utcdate} is provided.

The default template mappings (view with "hg kwdemo -d") can be replaced
with customized keywords and templates.
Again, run "hg kwdemo" to control the results of your config changes.

Before changing/disabling active keywords, run "hg kwshrink" to avoid
the risk of inadvertedly storing expanded keywords in the change history.

To force expansion after enabling it, or a configuration change, run
"hg kwexpand".

Expansions spanning more than one line and incremental expansions,
like CVS' $Log$, are not supported. A keyword template map
"Log = {desc}" expands to the first line of the changeset description.

Caveat: "hg import" fails if the patch context contains an active
        keyword. In that case run "hg kwshrink", and then reimport.
        Or, better, use bundle/unbundle to share changes.
'''

from mercurial import commands, cmdutil, context, fancyopts, filelog
from mercurial import localrepo, templater, util, hg
from mercurial.i18n import gettext as _
import getopt, re, shutil, sys, tempfile, time

# backwards compatibility hacks

try:
    # cmdutil.parse moves to dispatch._parse in 18a9fbb5cd78
    from mercurial import dispatch
    _parse = dispatch._parse
except ImportError:
    try:
        # commands.parse moves to cmdutil.parse in 0c61124ad877
        _parse = cmdutil.parse
    except AttributeError:
        _parse = commands.parse

try:
    # bail_if_changed moves from commands to cmdutil in 0c61124ad877
    bail_if_changed = cmdutil.bail_if_changed
except AttributeError:
    bail_if_changed = commands.bail_if_changed

# commands.parse/cmdutil.parse returned nothing for
# "hg diff --rev" before 88803a69b24a due to bug in fancyopts
def _fancyopts(args, options, state):
    '''Fixed fancyopts from 88803a69b24a.'''
    long = []
    short = ''
    map = {}
    dt = {}
    for s, l, d, c in options:
        pl = l.replace('-', '_')
        map['-'+s] = map['--'+l] = pl
        if isinstance(d, list):
            state[pl] = d[:]
        else:
            state[pl] = d
        dt[pl] = type(d)
        if (d is not None and d is not True and d is not False and
            not callable(d)):
            if s: s += ':'
            if l: l += '='
        if s: short = short + s
        if l: long.append(l)
    opts, args = getopt.getopt(args, short, long)
    for opt, arg in opts:
        if dt[map[opt]] is type(fancyopts): state[map[opt]](state, map[opt], arg)
        elif dt[map[opt]] is type(1): state[map[opt]] = int(arg)
        elif dt[map[opt]] is type(''): state[map[opt]] = arg
        elif dt[map[opt]] is type([]): state[map[opt]].append(arg)
        elif dt[map[opt]] is type(None): state[map[opt]] = True
        elif dt[map[opt]] is type(False): state[map[opt]] = True
    return args

fancyopts.fancyopts = _fancyopts


commands.optionalrepo += ' kwdemo'

def utcdate(date):
    '''Returns hgdate in cvs-like UTC format.'''
    return time.strftime('%Y/%m/%d %H:%M:%S', time.gmtime(date[0]))

class kwtemplater(object):
    '''
    Sets up keyword templates, corresponding keyword regex, and
    provides keyword substitution functions.
    '''
    templates = {
        'Revision': '{node|short}',
        'Author': '{author|user}',
        'Date': '{date|utcdate}',
        'RCSFile': '{file|basename},v',
        'Source': '{root}/{file},v',
        'Id': '{file|basename},v {node|short} {date|utcdate} {author|user}',
        'Header': '{root}/{file},v {node|short} {date|utcdate} {author|user}',
    }

    def __init__(self, ui, repo, expand, path='', node=None):
        self.ui = ui
        self.repo = repo
        self.t = expand or None
        self.path = path
        self.node = node

        kwmaps = self.ui.configitems('keywordmaps')
        if kwmaps:
            self.templates = {}
            for k, v in kwmaps:
                self.templates[k] = templater.parsestring(v, quoted=False)
        escaped = [re.escape(k) for k in self.templates.keys()]
        rawkeyword = r'\$(%s)(: [^$\n\r]*? )??\$'
        self.re_kw = re.compile(rawkeyword % '|'.join(escaped))
        if self.t:
            templater.common_filters['utcdate'] = utcdate
            self.t = self._changeset_templater()

    def _changeset_templater(self):
        '''Backwards compatible cmdutil.changeset_templater.'''
        try:
            return cmdutil.changeset_templater(self.ui, self.repo,
                                               False, '', False)
        except TypeError:
            # depending on hg rev changeset_templater has extra "brinfo" arg
            return cmdutil.changeset_templater(self.ui, self.repo,
                                               False, None, '', False)

    def _ctxnode(self, node):
        '''Obtains missing node from file context.'''
        if not self.node:
            c = context.filectx(self.repo, self.path, fileid=node)
            self.node = c.node()

    def _kwsub(self, mobj):
        '''Substitutes keyword using corresponding template.'''
        kw = mobj.group(1)
        self.t.use_template(self.templates[kw])
        self.ui.pushbuffer()
        self.t.show(changenode=self.node, root=self.repo.root, file=self.path)
        keywordsub = templater.firstline(self.ui.popbuffer())
        return '$%s: %s $' % (kw, keywordsub)

    def expand(self, node, data):
        '''Returns data with keywords expanded.'''
        if util.binary(data):
            return data
        self._ctxnode(node)
        return self.re_kw.sub(self._kwsub, data)

    def process(self, node, data):
        '''Returns a tuple: data, count.
        Count is number of keywords/keyword substitutions.
        Keywords in data are expanded, if templater was initialized.'''
        if util.binary(data):
            return data, None
        if self.t:
            self._ctxnode(node)
            return self.re_kw.subn(self._kwsub, data)
        return data, self.re_kw.search(data)

    def shrink(self, text):
        '''Returns text with all keyword substitutions removed.'''
        if util.binary(text):
            return text
        return self.re_kw.sub(r'$\1$', text)

    def overwrite(self, candidates, man, commit):
        '''Overwrites files in working directory if keywords are detected.
        Keywords are expanded if keyword templater is initialized,
        otherwise their substitution is removed.'''
        expand = self.t is not None
        action = ('shrinking', 'expanding')[expand]
        notify = (self.ui.note, self.ui.debug)[commit]
        overwritten = []
        for f in candidates:
            fp = self.repo.file(f, kwexp=expand, kwcnt=True)
            data, kwfound = fp.read(man[f])
            if kwfound:
                notify(_('overwriting %s %s keywords\n') % (f, action))
                try:
                    self.repo.wwrite(f, data, man.flags(f))
                except AttributeError:
                    # older versions want file descriptor as 3. optional arg
                    self.repo.wwrite(f, data)
                overwritten.append(f)
        if overwritten:
            try:
                for f in overwritten:
                    self.repo.dirstate.normal(f)
            except AttributeError:
                # older versions don't know about dirstate.normal
                self.repo.dirstate.update(overwritten, 'n')

class kwfilelog(filelog.filelog):
    '''
    Subclass of filelog to hook into its read, add, cmp methods.
    Keywords are "stored" unexpanded, and processed on reading.
    '''
    def __init__(self, opener, path, kwtemplater, kwcnt):
        super(kwfilelog, self).__init__(opener, path)
        self.kwtemplater = kwtemplater
        self.kwcnt = kwcnt

    def read(self, node):
        '''Passes data through kwemplater methods for
        either unconditional keyword expansion
        or counting of keywords and substitution.'''
        data = super(kwfilelog, self).read(node)
        if not self.kwcnt:
            return self.kwtemplater.expand(node, data)
        return self.kwtemplater.process(node, data)

    def add(self, text, meta, tr, link, p1=None, p2=None):
        '''Removes keyword substitutions when adding to filelog.'''
        text = self.kwtemplater.shrink(text)
        return super(kwfilelog, self).add(text, meta, tr, link, p1=p1, p2=p2)

    def cmp(self, node, text):
        '''Removes keyword substitutions for comparison.'''
        text = self.kwtemplater.shrink(text)
        if self.renamed(node):
            t2 = super(kwfilelog, self).read(node)
            return t2 != text
        return super(kwfilelog, self).cmp(node, text)

def _keywordmatcher(ui, repo):
    '''Collects include/exclude filename patterns for expansion
    candidates of current configuration. Returns filename matching
    function if include patterns exist, None otherwise.'''
    inc, exc = [], ['.hg*']
    for pat, opt in ui.configitems('keyword'):
        if opt != 'ignore':
            inc.append(pat)
        else:
            exc.append(pat)
    if inc:
        return util.matcher(repo.root, inc=inc, exc=exc)[1]
    return None

def _weedcandidates(man, kwfmatcher, candidates):
    '''Weeds out files that do not match keyword file matcher,
    are not tracked, or are links.'''
    files = man.keys()
    if candidates:
        return [f for f in candidates if kwfmatcher(f)
                and f in files and not man.linkf(f)]
    # kwexpand w/o args on all files in manifest
    return [f for f in files if kwfmatcher(f) and not man.linkf(f)]

def _overwrite(ui, repo, files, expand):
    '''Expands/shrinks keywords in working directory.'''
    wlock = lock = None
    try:
        wlock = repo.wlock()
        lock = repo.lock()
        bail_if_changed(repo)
        ctx = repo.changectx()
        if not ctx:
            raise hg.RepoError(_('no revision checked out'))
        kwfmatcher = _keywordmatcher(ui, repo)
        if kwfmatcher is None:
            ui.warn(_('no files configured for keyword expansion\n'))
            return
        man = ctx.manifest()
        files = _weedcandidates(man, kwfmatcher, files)
        if not files:
            ui.warn(_('files not configured for expansion or untracked\n'))
            return
        commit = False
        kwt = kwtemplater(ui, repo, expand, node=ctx.node())
        kwt.overwrite(files, man, commit)
    finally:
        del wlock, lock


def shrink(ui, repo, *args):
    '''revert expanded keywords in working directory

    run before changing/disabling active keywords
    or if you experience problems with "hg import" or "hg merge"
    '''
    expand = False
    _overwrite(ui, repo, args, expand)

def expand(ui, repo, *args):
    '''expand keywords in working directory

    run after (re)enabling keyword expansion
    '''
    expand = True
    _overwrite(ui, repo, args, expand)

def demo(ui, repo, *args, **opts):
    '''print [keywordmaps] configuration and an expansion example

    show current, custom, or default keyword template maps and their expansion

    extend current configuration by specifying maps as arguments
    and optionally by reading from an additional hgrc file

    override current keyword template maps with "default" option
    '''
    def _demostatus(stat):
        ui.status(_('\n\t%s\n') % stat)

    def _showitems(section, items):
        ui.write('[%s]\n' % section)
        for k, v in items:
            ui.write('%s = %s\n' % (k, v))

    msg = 'hg keyword config and expansion example'
    kwstatus = 'current'
    fn = 'demo.txt'
    branchname = 'demobranch'
    tmpdir = tempfile.mkdtemp('', 'kwdemo.')
    ui.note(_('creating temporary repo at %s\n') % tmpdir)
    repo = localrepo.localrepository(ui, path=tmpdir, create=True)
    ui.setconfig('keyword', fn, '')
    if args or opts['rcfile']:
        kwstatus = 'custom'
    if opts['rcfile']:
        ui.readconfig(opts['rcfile'])
    if opts['default']:
        kwstatus = 'default'
        kwmaps = kwtemplater.templates
        if ui.configitems('keywordmaps'):
            # override maps from optional rcfile
            for k, v in kwmaps.items():
                ui.setconfig('keywordmaps', k, v)
    elif args:
        # simulate hgrc parsing
        rcmaps = ['[keywordmaps]\n'] + [a + '\n' for a in args]
        fp = repo.opener('hgrc', 'w')
        fp.writelines(rcmaps)
        fp.close()
        ui.readconfig(repo.join('hgrc'))
    if not opts['default']:
        kwmaps = dict(ui.configitems('keywordmaps')) or kwtemplater.templates
    reposetup(ui, repo)
    for k, v in ui.configitems('extensions'):
        if k.endswith('keyword'):
            extension = '%s = %s' % (k, v)
            break
    _demostatus('config using %s keyword template maps' % kwstatus)
    ui.write('[extensions]\n%s\n' % extension)
    _showitems('keyword', ui.configitems('keyword'))
    _showitems('keywordmaps', kwmaps.items())
    keywords = '$' + '$\n$'.join(kwmaps.keys()) + '$\n'
    repo.wopener(fn, 'w').write(keywords)
    repo.add([fn])
    path = repo.wjoin(fn)
    ui.note(_('\n%s keywords written to %s:\n') % (kwstatus, path))
    ui.note(keywords)
    ui.note('\nhg -R "%s" branch "%s"\n' % (tmpdir, branchname))
    # silence branch command if not verbose
    quiet = ui.quiet
    verbose = ui.verbose
    ui.quiet = not verbose
    commands.branch(ui, repo, branchname)
    ui.quiet = quiet
    ui.note('hg -R "%s" ci -m "%s"\n' % (tmpdir, msg))
    repo.commit(text=msg)
    pathinfo = ('', ' in %s' % path)[ui.verbose]
    _demostatus('%s keywords expanded%s' % (kwstatus, pathinfo))
    ui.write(repo.wread(fn))
    ui.debug(_('\nremoving temporary repo %s\n') % tmpdir)
    shutil.rmtree(tmpdir)


def reposetup(ui, repo):
    '''Sets up repo as kwrepo for keyword substitution.
    Overrides file method to return kwfilelog instead of filelog
    if file matches user configuration.
    Wraps commit to overwrite configured files with updated
    keyword substitutions.
    This is done for local repos only, and only if there are
    files configured at all for keyword substitution.'''

    nokwcommands = ['add', 'addremove', 'bundle', 'clone', 'copy', 'export',
                    'grep', 'identify', 'incoming', 'init', 'outgoing', 'push',
                    'remove', 'rename', 'rollback']

    if not repo.local() or _parse(ui, sys.argv[1:])[0] in nokwcommands:
        return

    kwfmatcher = _keywordmatcher(ui, repo)
    if kwfmatcher is None:
        return

    class kwrepo(repo.__class__):
        def file(self, f, kwexp=True, kwcnt=False):
            if f[0] == '/':
                f = f[1:]
            if kwfmatcher(f):
                kwt = kwtemplater(ui, self, kwexp, path=f)
                return kwfilelog(self.sopener, f, kwt, kwcnt)
            return filelog.filelog(self.sopener, f)

        def _commit(self, files, text, user, date, match, force, lock, wlock,
                    force_editor, p1, p2, extra):
            '''Private commit wrapper for backwards compatibility.'''
            try:
                return super(kwrepo, self).commit(files=files, text=text,
                                                  user=user, date=date,
                                                  match=match, force=force,
                                                  lock=lock, wlock=wlock,
                                                  force_editor=force_editor,
                                                  p1=p1, p2=p2, extra=extra)
            except TypeError:
                return super(kwrepo, self).commit(files=files, text=text,
                                                  user=user, date=date,
                                                  match=match, force=force,
                                                  force_editor=force_editor,
                                                  p1=p1, p2=p2, extra=extra)

        def commit(self, files=None, text='', user=None, date=None,
                   match=util.always, force=False, lock=None, wlock=None,
                   force_editor=False, p1=None, p2=None, extra={}):
            # (w)lock arguments removed in 126f527b3ba3
            # so they are None or what was passed to commit
            # use private _(w)lock for deletion
            _lock = lock
            _wlock = wlock
            del wlock, lock
            try:
                if not _wlock:
                    _wlock = self.wlock()
                if not _lock:
                    _lock = self.lock()
                node = self._commit(files, text, user, date, match, force,
                                    _lock, _wlock, force_editor, p1, p2, extra)
                if node is not None:
                    cl = self.changelog.read(node)
                    mn = self.manifest.read(cl[0])
                    candidates = _weedcandidates(mn, kwfmatcher, cl[3])
                    if candidates:
                        expand = commit = True
                        kwt = kwtemplater(ui, self, expand, node=node)
                        kwt.overwrite(candidates, mn, commit)
                return node
            finally:
                del _wlock, _lock

    repo.__class__ = kwrepo


cmdtable = {
    'kwdemo':
        (demo,
         [('d', 'default', None, _('show default keyword template maps')),
          ('f', 'rcfile', [], _('read maps from RCFILE'))],
         _('hg kwdemo [-d] [-f RCFILE] [TEMPLATEMAP ...]')),
    'kwshrink': (shrink, [], _('hg kwshrink [NAME] ...')),
    'kwexpand': (expand, [], _('hg kwexpand [NAME] ...')),
}
